#ifndef __LIB_XP_H__
#define __LIB_XP_H__

//////////////////////////////////////////////////////////////////////////

#include "c4d.h"

namespace xparticles_lib
{

const LONG OXPEmitter = 1027133;
const LONG OXPDomain = 1033595;
const LONG MXPmaterial = 1027417;
const LONG VXPrender = 1030488;

// X-Particle 2.5 flags for compatibility

enum XPARTICLE_FLAGS_25
{
	XPARTICLE_FLAGS_25_0				= 0,
	XPARTICLE_FLAGS_25_ALIVE		= (1 << 0),
	XPARTICLE_FLAGS_25_BORN		= (1 << 1),
	XPARTICLE_FLAGS_25_DIE			= (1 << 2),
	XPARTICLE_FLAGS_25_GROUP_CHANGED = (1 << 3),
	XPARTICLE_FLAGS_25_NEW			= (1 << 4),
} ENUM_END_FLAGS(XPARTICLE_FLAGS_25);

enum XPARTICLE_BITS
{
	XPARTICLE_BITS_0				= 0,
	XPARTICLE_BITS_TRAIL = 1,
	XPARTICLE_BITS_USEGRAD = 2,
	XPARTICLE_BITS_FREEZE = 4,
	XPARTICLE_BITS_STICK = 8,
	XPARTICLE_BITS_ISSTUCK = 16,
	XPARTICLE_BITS_COLLIDED = 32,
	XPARTICLE_BITS_STICKSOURCE = 64,
	XPARTICLE_BITS_TARGETMODE = 256,
	XPARTICLE_BITS_FREEZE_SPIN = 512,
	XPARTICLE_BITS_FREEZE_SCALE = 1024,
	XPARTICLE_BITS_TP_DISCON = 2048,
	XPARTICLE_BITS_ROTATED = 4096,
	XPARTICLE_BITS_DONESPAWN = 8192,
	XPARTICLE_BITS_DOSTICK = 16384,
	XPARTICLE_BITS_DOFREEZE_MOVE = 32768,
	XPARTICLE_BITS_DOFREEZE_SPIN = 65536,
	XPARTICLE_BITS_DOFREEZE_SCALE = 131072,
	XPARTICLE_BITS_CUTTRAIL = 262144,
	XPARTICLE_BITS_KILLZEROTRAIL = 524288,
	XPARTICLE_BITS_NOSTUCKACTION = (1<<20),
	XPARTICLE_BITS_DIRECTKICK = (1<<21),
	XPARTICLE_BITS_INVISIBLE = (1<<22),
	XPARTICLE_BITS_COLLISION_SKIP = (1<<23),
	XPARTICLE_BITS_COLLISION_STEP_SKIP = (1<<24),
	XPARTICLE_BITS_PPCOLLISION = (1<<25),
	XPARTICLE_BITS_PPOLYCOLLISION = (1<<26),
	XPARTICLE_BITS_POLY_PRIMECOLLISION = (1<<27),
	XPARTICLE_BITS_PP_PRIMECOLLISION = (1<<28),
	XPARTICLE_BITS_DEFORMED = (1<<29),
} ENUM_END_FLAGS(XPARTICLE_BITS);

// end 2.5 compatibility

enum XPARTICLE_FLAGS
{
	XPARTICLE_FLAGS_0				= 0,
	XPARTICLE_FLAGS_TRAIL						= (1 << 0),
	XPARTICLE_FLAGS_USEGRAD					= (1 << 1),
	XPARTICLE_FLAGS_FREEZE					= (1 << 2),
	XPARTICLE_FLAGS_STICK						= (1 << 3),
	XPARTICLE_FLAGS_ISSTUCK					= (1 << 4),
	XPARTICLE_FLAGS_COLLIDED					= (1 << 5),
	XPARTICLE_FLAGS_STICKSOURCE				= (1 << 6),
	XPARTICLE_FLAGS_NEWBRANCH					= (1 << 7),
	XPARTICLE_FLAGS_RESAMPLE					= (1 << 8),
	XPARTICLE_FLAGS_FREEZE_SPIN				= (1 << 9),
	XPARTICLE_FLAGS_FREEZE_SCALE				= (1 << 10),
	XPARTICLE_FLAGS_TP_DISCON					= (1 << 11),
	XPARTICLE_FLAGS_ROTATED					= (1 << 12),
	XPARTICLE_FLAGS_DONESPAWN					= (1 << 13),
	XPARTICLE_FLAGS_DOSTICK					= (1 << 14),
	XPARTICLE_FLAGS_ALIVE		= (1 << 15),
	XPARTICLE_FLAGS_BORN		= (1 << 16),
	XPARTICLE_FLAGS_DIE			= (1 << 17),
	XPARTICLE_FLAGS_GROUP_CHANGED = (1 << 18),
	XPARTICLE_FLAGS_NEW			= (1 << 19),
	XPARTICLE_FLAGS_NOSTUCKACTION				= (1 << 20),
	XPARTICLE_FLAGS_INVISIBLE					= (1 << 21),
	XPARTICLE_FLAGS_SKIP_FINAL			= (1 << 22),
	XPARTICLE_FLAGS_PPCOLLISION				= (1 << 24),
	XPARTICLE_FLAGS_PPOLYCOLLISION			= (1 << 25),
	XPARTICLE_FLAGS_POLY_PRIMECOLLISION		= (1 << 26),
	XPARTICLE_FLAGS_PP_PRIMECOLLISION			= (1 << 27),
	XPARTICLE_FLAGS_DEFORMED					= (1 << 28),
	XPARTICLE_FLAGS_STREAMER					= (1 << 29),
	XPARTICLE_FLAGS_MARKED_DIE = (1 << 30),
	XPARTICLE_FLAGS_FILL		= (1 << 31),
} ENUM_END_FLAGS(XPARTICLE_FLAGS);

class XParticleConstraint
{
	XParticleConstraint();
	~XParticleConstraint();

public:

	Real GetLength();
	LONG GetIndex();
	LONG GetFlags();
	LONG GetID();

	void SetLength(Real len);
	void SetIndex(LONG index);
	void SetFlags(LONG flags);
	void SetID(LONG id);
};

class XParticleColliderConstraint
{
	XParticleColliderConstraint();
	~XParticleColliderConstraint();

public:

	Real GetLength();
	LONG GetIndex();
	LONG GetFlags();
	LONG GetID();
	LULONG GetColliderID();
	void GetPositionRS(Real &r, Real &s);
	Vector GetLastPosition();

	void SetLength(Real len);
	void SetIndex(LONG index);
	void SetFlags(LONG flags);
	void SetID(LONG id);
	void SetColliderID(LULONG id);
	void SetPositionRS(Real r, Real s);
	void SetLastPosition(const Vector &pos);
};

Vector MixRS(Real r, Real s, Real ra, Real rb, Real rc, Real rd, Bool quad);
Vector MixRS(Real r, Real s, const Vector &va, const Vector &vb, const Vector &vc, const Vector &vd, Bool quad);
void GetRS(Real &r, Real &s, const Vector &p, const Vector &va, const Vector &vb, const Vector &vc, const Vector &vd, Bool quad);

class XParticle
{
private:

	XParticle();
	~XParticle();

public:

	Vector GetPosition();
	Vector GetVelocity();
	Real GetTime();
	Real GetLife();
	Vector GetColor();
	Real GetRadius();
	Real GetMass();
	Real GetSpeed();
	LONG GetGroupID();
	LONG GetID();
	XPARTICLE_FLAGS GetFlags();
	Real GetDensity();			// only valid with fluids enabled
	LONG GetDisplay();			// get the particle display type (defined in the emitter resource, res folder)
	XPARTICLE_BITS GetBits();
	Matrix GetMatrix();
	Vector GetRotation();
	Real GetDistanceTravelled();
	Vector GetSpin();
	Bool IsSpinEnabled();
	Real GetTemperature();
	Real GetFuel();
	Real GetFire();
	Real GetSmoke();
	Real GetExpansion();
	Vector GetUVW();

	void SetPosition(const Vector &p);
	void SetVelocity(const Vector &v);
	void SetTime(Real time);
	void SetLife(Real life);
	void SetColor(const Vector &col);
	void SetRadius(Real r);
	void SetMass(Real m);
	void SetSpeed(Real s);
	void SetGroupID(LONG id);
	void SetID(LONG id);
	void SetFlags(XPARTICLE_FLAGS flags);	// particles can not be created by changing their flags, please use CreateParticle, flags should generally not be changed
	void SetDensity(Real density);
	void SetDisplay(LONG type);
	void SetBits(XPARTICLE_BITS bits);
	void SetMatrix(const Matrix &m);
	void SetRotation(const Vector &rot);	// HPB
	void SetDistanceTravelled(Real dist);
	void SetSpin(const Vector &spin);
	void SetSpinEnabled(Bool enable);
	void SetTemperature(Real temp);
	void SetFuel(Real fuel);
	void SetFire(Real fire);
	void SetSmoke(Real smoke);
	void SetExpansion(Real expansion);
	void SetUVW(const Vector &uvw);

	GeData *GetCustomData(LONG index, LONG *id = NULL, String *name = NULL);
	LONG GetCustomDataCount();
	LONG AddCustomData(LONG id, const String &name, const GeData &data);
	void SetCustomData(LONG index, const GeData &data);
	LONG FindCustomData(LONG id, const String &name);	// find custom data matching id/name (id can be -1 if only the name is to be checked, name can be String() if only the id should be checked)
	Bool RemoveCustomData(LONG index);

	LONG GetBirthConstraintsCount();
	LONG GetDistanceConstraintsCount();
	LONG GetParticleConstraintsCount();
	LONG GetColliderConstraintsCount();

	XParticleConstraint *GetBirthConstraint(LONG index);
	XParticleConstraint *GetDistanceConstraint(LONG index);
	XParticleConstraint *GetParticleConstraint(LONG index);
	XParticleColliderConstraint *GetColliderConstraint(LONG index);

	XParticleConstraint *AddBirthConstraint();
	XParticleConstraint *AddDistanceConstraint();
	XParticleConstraint *AddParticleConstraint();
	XParticleConstraint *AddColliderConstraint();

	void RemoveBirthConstraint(LONG index);
	void RemoveDistanceConstraint(LONG index);
	void RemoveParticleConstraint(LONG index);
	void RemoveColliderConstraint(LONG index);

	LONG GetIndex();					// get the interal array index of this particle, iterating should be done using GetFirstAlive() and GetNext()
	XParticle *GetNext();			// get the next used (live) particle
};

};	// xparticles_lib

class XPVideoPost : public BaseVideoPost
{
private:

	XPVideoPost();
	~XPVideoPost();

public:

	static XPVideoPost *Alloc()
	{
		return (XPVideoPost*)BaseVideoPost::Alloc(xparticles_lib::VXPrender);
	}

	static void Free(XPVideoPost *&xpVp)
	{
		BaseVideoPost *vp = xpVp; 
		BaseVideoPost::Free(vp);
		xpVp = NULL;
	}

	void SkipRender();
};

// to use the X-Particle Material Sampling the material must have been initialized (InitRender)

class XPMaterial : public BaseMaterial
{
private:

	XPMaterial();
	~XPMaterial();

public:

	static XPMaterial *Alloc()
	{
		return (XPMaterial*)BaseMaterial::Alloc(xparticles_lib::MXPmaterial);
	}

	static void Free(XPMaterial *&xpMat)
	{
		BaseMaterial *mat = xpMat; 
		BaseMaterial::Free(mat);
		xpMat = NULL;
	}

	Real SampleSize(xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render);
	Vector SampleColor(xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render, Real radius);
	Real SampleTrans(xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render, Real radius);
};

// iterate all particles using GetFirstAlive() followed by GetNext() from the particle
//
// xparticles_lib::XParticle *particle = emitter->GetFirstAlive();
// while (particle != nullptr)
// {
//   // do something
//   particle = particle->GetNext();
// }
//

class XPEmitterObject : public BaseObject
{
private:

	XPEmitterObject();
	~XPEmitterObject();

public:

	static XPEmitterObject *Alloc()
	{
		return (XPEmitterObject*)BaseObject::Alloc(xparticles_lib::OXPEmitter);
	}

	static void Free(XPEmitterObject *&xpObject)
	{
		BaseObject *op = xpObject; 
		BaseObject::Free(op);
		xpObject = NULL;
	}

	LONG GetParticleCount();		// total particle count
	LONG GetAliveCount();				// count of 'in use' (live) particles
	LONG GetDeadCount();				// unused (dead) particles
	LONG GetMaxCount();

	xparticles_lib::XParticle *GetParticle(LONG index);		// get a particle at 'index', NOTE: iterating the particles using this function is slower than using GetFirstAlive()
	xparticles_lib::XParticle *GetFirstAlive();						// get the first live particle, iterate through all live particles from this with GetNext()
	xparticles_lib::XParticle *GetLastAlive();

	xparticles_lib::XParticle *CreateParticle();				// create a live particle, position/velocity/life etc. will need setting
	Bool SetParticleCount(LONG count);									// set the number of live particles, each particle will need init
	Bool SetMaxCount(LONG count);

	LONG GetFlags();																		// currently private
	void SetFlags(LONG flags);													// currently private
};

String XPGetVersion();

class XPC4DParticleModifierMsg
{
public:

	XPC4DParticleModifierMsg()
	{
		_emitter = nullptr;
	}

	XPEmitterObject *_emitter;
};

static const LONG MSG_XPARTICLES_ISMODIFIER = 1029569;
static const LONG MSG_XPARTICLES_C4DMODIFY = 1032239;

enum XP_MODIFIERMSG_FLAG
{
	XP_MODIFIERMSG_FLAG_0 = 0,
	XP_MODIFIERMSG_FLAG_PLUGIN_MODIFIER = (1 << 1),
	XP_MODIFIERMSG_FLAG_PLUGIN_THREAD = (1 << 2),
} ENUM_END_FLAGS(XP_MODIFIERMSG_FLAG);

class XPModifierMsgData
{
public:

	XPModifierMsgData()
	{
		_flags = XP_MODIFIERMSG_FLAG_0;
	}

	XP_MODIFIERMSG_FLAG _flags;
};

class XPModifierObject : public ObjectData
{
	INSTANCEOF(XPModifierObject, ObjectData)

public:

	virtual Bool Message(GeListNode *node, LONG type, void *data)
	{
		if (type == MSG_XPARTICLES_ISMODIFIER && data != nullptr)
		{
			XPModifierMsgData *msgData = (XPModifierMsgData*)data;
			msgData->_flags |= XP_MODIFIERMSG_FLAG_PLUGIN_MODIFIER | XP_MODIFIERMSG_FLAG_PLUGIN_THREAD;
		}

		return SUPER::Message(node, type, data);
	}

	// return FALSE if anything went wrong and this modifier should be skipped
	virtual Bool InitModifyXP(BaseDocument *doc, BaseObject *modObject, BaseContainer *modData, XPEmitterObject *emitter, Real timeDiff)
	{
		return TRUE;
	}
	
	// always called if InitModifyXP returned TRUE
	virtual void FlushModifyXP()
	{
	}

	// startIndex/endIndex is NOTOK when not using the threaded (XP_MODIFIERMSG_FLAG_PLUGIN_THREAD) flag
	// return FALSE to stop (if Message set XP_MODIFIERMSG_FLAG_PLUGIN_THREAD)
	virtual Bool ModifyXP(BaseDocument *doc, BaseObject *modObject, BaseContainer *modData, XPEmitterObject *emitter, LONG startIndex, LONG endIndex, Real timeDiff)
	{
		return TRUE;
	}
	/* Threaded particle can be iterated as below
	Bool Gravitation::ModifyXP(BaseDocument *doc, BaseObject *modObject, BaseContainer *modData, XPEmitterObject *emitter, LONG startIndex, LONG endIndex, Real timeDiff)
	{
		xparticles_lib::XParticle *particle = emitter->GetParticle(startIndex);

		while (particle != nullptr && particle->GetIndex() <= endIndex)
		{
			particle = particle->GetNext();
		}

		return TRUE;
	}
	*/
};

//////////////////////////////////////////////////////////////////////////

class XPDomainObject : public BaseObject
{
private:

	XPDomainObject();
	~XPDomainObject();

public:

	static XPDomainObject *Alloc()
	{
		return (XPDomainObject*)BaseObject::Alloc(xparticles_lib::OXPEmitter);
	}

	static void Free(XPDomainObject *&xpObject)
	{
		BaseObject *op = xpObject; 
		BaseObject::Free(op);
		xpObject = NULL;
	}

	XPEmitterObject *GetGasEmitter();

	LLONG GetWidth();
	LLONG GetHeight();
	LLONG GetDepth();

	void GetVoxelXYZ(const Vector &p, LLONG &x, LLONG &y, LLONG &z);
	Vector GetVoxelCenter(LLONG x, LLONG y, LLONG z);

	Vector GetVelocity(const Vector &p);
	Vector GetVelocity(LLONG x, LLONG y, LLONG z);
	Real GetSmoke(const Vector &p);
	Real GetSmoke(LLONG x, LLONG y, LLONG z);
	Real GetFire(const Vector &p);
	Real GetFire(LLONG x, LLONG y, LLONG z);
	Real GetFuel(const Vector &p);
	Real GetFuel(LLONG x, LLONG y, LLONG z);
	Real GetTemperature(const Vector &p);
	Real GetTemperature(LLONG x, LLONG y, LLONG z);
	Vector GetUVW(const Vector &p);
	Vector GetUVW(LLONG x, LLONG y, LLONG z);
};

//////////////////////////////////////////////////////////////////////////

// private library definitions

#define LIBRARY_XPARTICLES		1029633

struct XParticlesLibrary : public C4DLibrary
{
	LONG (*xpemitGetParticleU)(XPEmitterObject *xpObject);
	LONG (*xpemitGetAliveCountP)(XPEmitterObject *xpObject);
	LONG (*xpemitGetDeadCount)(XPEmitterObject *xpObject);
	xparticles_lib::XParticle *(*xpemitGetParticleI)(XPEmitterObject *xpObject, LONG index);
	xparticles_lib::XParticle *(*xpemitCreateParticle)(XPEmitterObject *xpObject);
	Bool (*xpemitSetParticleCount)(XPEmitterObject *xpObject, LONG count);
	LONG (*xpemitGetFlags)(XPEmitterObject *xpObject);
	void (*xpemitSetFlags)(XPEmitterObject *xpObject, LONG flags);

	Vector (*xppartGetVelocity)(xparticles_lib::XParticle *xpParticle);
	Vector (*xppartGetPosition)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetLife)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetTime)(xparticles_lib::XParticle *xpParticle);
	Vector (*xppartGetColor)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetMass)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetRadius)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetSpeed)(xparticles_lib::XParticle *xpParticle);
	LONG (*xppartGetGroupID)(xparticles_lib::XParticle *xpParticle);
	LONG (*xppartGetID)(xparticles_lib::XParticle *xpParticle);
	xparticles_lib::XPARTICLE_FLAGS (*xppartGetFlags25)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetDensity)(xparticles_lib::XParticle *xpParticle);

	void (*xppartSetPosition)(xparticles_lib::XParticle *xpParticle, const Vector &p);
	void (*xppartSetVelocity)(xparticles_lib::XParticle *xpParticle, const Vector &v);
	void (*xppartSetTime)(xparticles_lib::XParticle *xpParticle, Real time);
	void (*xppartSetLife)(xparticles_lib::XParticle *xpParticle, Real life);
	void (*xppartSetColor)(xparticles_lib::XParticle *xpParticle, const Vector &col);
	void (*xppartSetRadius)(xparticles_lib::XParticle *xpParticle, Real r);
	void (*xppartSetMass)(xparticles_lib::XParticle *xpParticle, Real m);
	void (*xppartSetSpeed)(xparticles_lib::XParticle *xpParticle, Real s);
	void (*xppartSetGroupID)(xparticles_lib::XParticle *xpParticle, LONG id);
	void (*xppartSetID)(xparticles_lib::XParticle *xpParticle, LONG id);
	void (*xppartSetFlags25)(xparticles_lib::XParticle *xpParticle, xparticles_lib::XPARTICLE_FLAGS flags);
	void (*xppartSetDensity)(xparticles_lib::XParticle *xpParticle, Real density);

	LONG (*xppartGetIndex)(xparticles_lib::XParticle *xpParticle);
	xparticles_lib::XParticle *(*xppartGetNextO)(xparticles_lib::XParticle *xpParticle);

	LONG (*xppartGetDisplay)(xparticles_lib::XParticle *xpParticle);
	void (*xppartSetDisplay)(xparticles_lib::XParticle *xpParticle, LONG type);

	xparticles_lib::XParticle *(*xpemitGetFirstAliveQ)(XPEmitterObject *xpObject);
	xparticles_lib::XParticle *(*xpemitGetLastAliveZ)(XPEmitterObject *xpObject);

	Real (*xpvpSampleSize)(XPMaterial *xpMaterial, xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render);
	Vector (*xpvpSampleColor)(XPMaterial *xpMaterial, xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render, Real radius);
	Real (*xpvpSampleTrans)(XPMaterial *xpMaterial, xparticles_lib::XParticle *xpParticle, const Vector &p, VolumeData *vd, XPVideoPost *render, Real radius);

	GeData *(*xppartGetCustomData)(xparticles_lib::XParticle *xpParticle, LONG index, LONG *id, String *name);
	LONG (*xppartAddCustomData)(xparticles_lib::XParticle *xpParticle, LONG id, const String &name, const GeData &data);
	void (*xppartSetCustomData)(xparticles_lib::XParticle *xpParticle, LONG index, const GeData &data);
	LONG (*xppartFindCustomData)(xparticles_lib::XParticle *xpParticle, LONG id, const String &name);
	LONG (*xppartGetCustomDataCount)(xparticles_lib::XParticle *xpParticle);

	LONG (*xpemitGetParticleCount)(XPEmitterObject *xpObject);
	xparticles_lib::XParticle *(*xpemitGetParticle)(XPEmitterObject *xpObject, LONG index);
	LONG (*xpemitGetAliveCount)(XPEmitterObject *xpObject);

	xparticles_lib::XParticle *(*xpemitGetFirstAlive)(XPEmitterObject *xpObject);
	xparticles_lib::XParticle *(*xpemitGetLastAlive)(XPEmitterObject *xpObject);

	xparticles_lib::XParticle *(*xppartGetNext)(xparticles_lib::XParticle *xpParticle);

	String (*xpGetVersion)();

	xparticles_lib::XPARTICLE_BITS (*xppartGetBits)(xparticles_lib::XParticle *xpParticle);
	Matrix (*xppartGetMatrix)(xparticles_lib::XParticle *xpParticle);
	Vector (*xppartGetRotation)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetDistanceTravelled)(xparticles_lib::XParticle *xpParticle);

	Bool (*xppartRemoveCustomData)(xparticles_lib::XParticle *xpParticle, LONG index);

	void (*xppartSetBits)(xparticles_lib::XParticle *xpParticle, xparticles_lib::XPARTICLE_BITS bits);
	void (*xppartSetMatrix)(xparticles_lib::XParticle *xpParticle, const Matrix &m);
	void (*xppartSetRotation)(xparticles_lib::XParticle *xpParticle, const Vector &rot);
	void (*xppartSetDistanceTravelled)(xparticles_lib::XParticle *xpParticle, Real dist);

	LONG (*xppartGetBirthConstraintsCount)(xparticles_lib::XParticle *xpParticle);
	LONG (*xppartGetDistanceConstraintsCount)(xparticles_lib::XParticle *xpParticle);
	LONG (*xppartGetParticleConstraintsCount)(xparticles_lib::XParticle *xpParticle);
	LONG (*xppartGetColliderConstraintsCount)(xparticles_lib::XParticle *xpParticle);

	xparticles_lib::XParticleConstraint *(*xppartGetBirthConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	xparticles_lib::XParticleConstraint *(*xppartGetDistanceConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	xparticles_lib::XParticleConstraint *(*xppartGetParticleConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	xparticles_lib::XParticleColliderConstraint *(*xppartGetColliderConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);

	xparticles_lib::XParticleConstraint *(*xppartAddBirthConstraint)(xparticles_lib::XParticle *xpParticle);
	xparticles_lib::XParticleConstraint *(*xppartAddDistanceConstraint)(xparticles_lib::XParticle *xpParticle);
	xparticles_lib::XParticleConstraint *(*xppartAddParticleConstraint)(xparticles_lib::XParticle *xpParticle);
	xparticles_lib::XParticleConstraint *(*xppartAddColliderConstraint)(xparticles_lib::XParticle *xpParticle);

	void (*xppartRemoveBirthConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	void (*xppartRemoveDistanceConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	void (*xppartRemoveParticleConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);
	void (*xppartRemoveColliderConstraint)(xparticles_lib::XParticle *xpParticle, LONG index);

	Real (*xpconstraintGetLength)(xparticles_lib::XParticleConstraint *xpConstraint);
	LONG (*xpconstraintGetIndex)(xparticles_lib::XParticleConstraint *xpConstraint);
	LONG (*xpconstraintGetFlags)(xparticles_lib::XParticleConstraint *xpConstraint);
	LONG (*xpconstraintGetID)(xparticles_lib::XParticleConstraint *xpConstraint);

	void (*xpconstraintSetLength)(xparticles_lib::XParticleConstraint *xpConstraint, Real len);
	void (*xpconstraintSetIndex)(xparticles_lib::XParticleConstraint *xpConstraint, LONG index);
	void (*xpconstraintSetFlags)(xparticles_lib::XParticleConstraint *xpConstraint, LONG flags);
	void (*xpconstraintSetID)(xparticles_lib::XParticleConstraint *xpConstraint, LONG id);

	Real (*xpcollconstraintGetLength)(xparticles_lib::XParticleColliderConstraint *xpConstraint);
	LONG (*xpcollconstraintGetIndex)(xparticles_lib::XParticleColliderConstraint *xpConstraint);
	LONG (*xpcollconstraintGetFlags)(xparticles_lib::XParticleColliderConstraint *xpConstraint);
	LONG (*xpcollconstraintGetID)(xparticles_lib::XParticleColliderConstraint *xpConstraint);
	LULONG (*xpcollconstraintGetColliderID)(xparticles_lib::XParticleColliderConstraint *xpConstraint);
	void (*xpcollconstraintGetPositionRS)(xparticles_lib::XParticleColliderConstraint *xpConstraint, Real &r, Real &s);
	Vector (*xpcollconstraintGetLastPosition)(xparticles_lib::XParticleColliderConstraint *xpConstraint);

	void (*xpcollconstraintSetLength)(xparticles_lib::XParticleColliderConstraint *xpConstraint, Real len);
	void (*xpcollconstraintSetIndex)(xparticles_lib::XParticleColliderConstraint *xpConstraint, LONG index);
	void (*xpcollconstraintSetFlags)(xparticles_lib::XParticleColliderConstraint *xpConstraint, LONG flags);
	void (*xpcollconstraintSetID)(xparticles_lib::XParticleColliderConstraint *xpConstraint, LONG id);
	void (*xpcollconstraintSetColliderID)(xparticles_lib::XParticleColliderConstraint *xpConstraint, LULONG id);
	void (*xpcollconstraintSetPositionRS)(xparticles_lib::XParticleColliderConstraint *xpConstraint, Real r, Real s);
	void (*xpcollconstraintSetLastPosition)(xparticles_lib::XParticleColliderConstraint *xpConstraint, const Vector &pos);

	Vector (*xpMixRSreal)(Real r, Real s, Real ra, Real rb, Real rc, Real rd, Bool quad);
	Vector (*xpMixRSvector)(Real r, Real s, const Vector &va, const Vector &vb, const Vector &vc, const Vector &vd, Bool quad);
	void (*xpGetRS)(Real &r, Real &s, const Vector &p, const Vector &va, const Vector &vb, const Vector &vc, const Vector &vd, Bool quad);

	Vector (*xppartGetSpin)(xparticles_lib::XParticle *xpParticle);
	Bool (*xppartIsSpinEnabled)(xparticles_lib::XParticle *xpParticle);
	void (*xppartSetSpin)(xparticles_lib::XParticle *xpParticle, const Vector &spin);
	void (*xppartSetSpinEnabled)(xparticles_lib::XParticle *xpParticle, Bool enable);

	LONG (*xpemitGetMaxCount)(XPEmitterObject *xpObject);
	Bool (*xpemitSetMaxCount)(XPEmitterObject *xpObject, LONG count);

	void (*xpvpSkipRender)(XPVideoPost *xpVP);

	xparticles_lib::XPARTICLE_FLAGS (*xppartGetFlags)(xparticles_lib::XParticle *xpParticle);
	void (*xppartSetFlags)(xparticles_lib::XParticle *xpParticle, xparticles_lib::XPARTICLE_FLAGS flags);

	void (*xppartSetTemperature)(xparticles_lib::XParticle *xpParticle, Real temp);
	void (*xppartSetFuel)(xparticles_lib::XParticle *xpParticle, Real fuel);
	void (*xppartSetFire)(xparticles_lib::XParticle *xpParticle, Real fire);
	void (*xppartSetSmoke)(xparticles_lib::XParticle *xpParticle, Real smoke);
	void (*xppartSetExpansion)(xparticles_lib::XParticle *xpParticle, Real expansion);
	Real (*xppartGetTemperature)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetFuel)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetFire)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetSmoke)(xparticles_lib::XParticle *xpParticle);
	Real (*xppartGetExpansion)(xparticles_lib::XParticle *xpParticle);
	void (*xppartSetUVW)(xparticles_lib::XParticle *xpParticle, const Vector &uvw);
	Vector (*xppartGetUVW)(xparticles_lib::XParticle *xpParticle);

	XPEmitterObject *(*xpdomainGetGasEmitter)(XPDomainObject *xpDomain);

	LLONG (*xpdomainGetWidth)(XPDomainObject *xpDomain);
	LLONG (*xpdomainGetHeight)(XPDomainObject *xpDomain);
	LLONG (*xpdomainGetDepth)(XPDomainObject *xpDomain);

	void (*xpdomainGetVoxelXYZ)(XPDomainObject *xpDomain, const Vector &p, LLONG &x, LLONG &y, LLONG &z);
	Vector (*xpdomainGetVoxelCenter)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);

	Vector(*xpdomainGetVelocity)(XPDomainObject *xpDomain, const Vector &p);
	Vector (*xpdomainGetVelocityXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
	Real (*xpdomainGetSmoke)(XPDomainObject *xpDomain, const Vector &p);
	Real (*xpdomainGetSmokeXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
	Real (*xpdomainGetFire)(XPDomainObject *xpDomain, const Vector &p);
	Real (*xpdomainGetFireXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
	Real (*xpdomainGetFuel)(XPDomainObject *xpDomain, const Vector &p);
	Real (*xpdomainGetFuelXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
	Real (*xpdomainGetTemperature)(XPDomainObject *xpDomain, const Vector &p);
	Real (*xpdomainGetTemperatureXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
	Vector (*xpdomainGetUVW)(XPDomainObject *xpDomain, const Vector &p);
	Vector (*xpdomainGetUVWXYZ)(XPDomainObject *xpDomain, LLONG x, LLONG y, LLONG z);
};

//////////////////////////////////////////////////////////////////////////

#endif	// __LIB_XP_H__
